/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj;

import java.util.concurrent.ExecutorService;

/**
 * A process is an instance of a {@link Program}.
 * 
 * @see Programs#as(Object)
 * 
 * @author Atos Worldline
 * 
 * @param <E>
 *            the type annotated with {@link Program}
 */
public interface Process<E> {

	public enum State {

		/**
		 * This is the initial state of a {@link Process}.
		 * 
		 * Transition to {@link #RUNNING} when the process is being executed.
		 * 
		 */
		CREATED,

		/**
		 * The program is currently running meaning that some procedures are
		 * currently being executed or new procedures will be launched.
		 * 
		 * <p>
		 * Transition to {@link #SUSPENDED} when {@link Process#suspend()} is
		 * triggered.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #COMPLETED} when all procedures are completed.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #ABORTED} when {@link Process#abort()} is
		 * triggered.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #TERMINATED} when {@link Process#terminate()} is
		 * triggered.
		 * </p>
		 * 
		 */
		RUNNING,

		/**
		 * The execution of the process has been suspended. No more procedures
		 * will be launched as long as {@link Process#resume()} has not been
		 * called. Note that already running procedures won't be suspended.
		 * <p>
		 * Transition to {@link #RUNNING} when {@link Process#resume()} is
		 * triggered.
		 * </p>
		 * 
		 */
		SUSPENDED,

		/**
		 * This is an exit state of a process. All running procedures have been
		 * canceled and no new procedures will be launched.
		 */
		ABORTED,

		/**
		 * This is an exit state of a process. All running procedures have been
		 * completed and no new procedures will be launched.
		 */
		TERMINATED,

		/**
		 * This is the default exit condition of a process. All procedures have
		 * been executed.
		 */
		COMPLETED;

	}

	/**
	 * Return the context bound to the process.
	 * 
	 * @return
	 */
	public E context();

	/**
	 * Execute the process with the current thread.
	 * 
	 */
	public Process<E> execute();

	/**
	 * Execute the process with a specific executor service.
	 * 
	 * @param service
	 */
	public Process<E> execute(ExecutorService service);

	/**
	 * Abort the process.
	 * 
	 */
	public Process<E> abort();

	/**
	 * Terminate the process.
	 */
	public Process<E> terminate();

	/**
	 * Suspend the execution of the process.
	 */
	public Process<E> suspend();

	/**
	 * Resume the execution of the process.
	 */
	public Process<E> resume();

	/**
	 * Wait for the completion of the process.
	 */
	public Process<E> join();

	/**
	 * Return the current state of the process.
	 * 
	 * @return
	 */
	public State getState();
}