/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj.internal.kernel;

/**
 * A call is an instance of an {@link KProcedure}.
 * 
 * <h1>Subclasses contract</h1>
 * <p>
 * Subclasses must inform when they {@link #started() start} and when they
 * {@link #completed() complete}.
 * 
 * @see KProcedure#newCall(KProcess)
 * 
 * @author Atos Worldline
 */
public class KCall {

	public enum State {

		/**
		 * The {@link KCall} has been created.
		 */
		CREATED {

			@Override
			void enter(KCall call) {
				call.doEnter();
				call.setState(PENDING);
			}
		},

		/**
		 * The {@link KCall} is ready to be started.
		 */
		PENDING {

			@Override
			void start(KCall call) {
				call.doStart();
				call.setState(RUNNING);
			}
		},

		/**
		 * The {@link KCall} is being executed.
		 */
		RUNNING {

			@Override
			void complete(KCall call) {
				call.doComplete();
				call.setState(COMPLETED);
			}

			@Override
			void cancel(KCall call) {
				call.doCancel();
				call.setState(CANCELED);
			}
		},

		COMPLETED {
			
			@Override
			void succeed(KCall call) {
				call.doSucceed();
				call.setState(SUCCEEDED);
			}
			
			@Override
			void discard(KCall call) {
				call.doDiscard();
				call.setState(DISCARDED);
			}
			
			@Override
			void fail(KCall call) {
				call.doFail();
				call.setState(FAILED);
			}

		},

		/**
		 * The {@link KCall} has been canceled.
		 */
		CANCELED,

		/**
		 * The {@link KCall} has completed but with exception.
		 */
		FAILED,

		/**
		 * The {@link KCall} has completed successfully.
		 */
		SUCCEEDED {

			@Override
			void exit(KCall call) {
				call.doSucceed();
			}
		},

		/**
		 * 
		 */
		DISCARDED;

		void enter(KCall call) {
		}

		void start(KCall call) {
		}

		void complete(KCall call) {
		}

		void cancel(KCall call) {
		}

		void exit(KCall call) {
		}

		void fail(KCall call) {
		}

		void succeed(KCall call) {
		}

		void discard(KCall call) {
		}

	}

	/**
	 * The procedure that created this instance.
	 */
	private final KProcedure procedure;

	/**
	 * The process that owns this instance.
	 */
	private final KProcess process;

	/**
	 * KTask level context.
	 * 
	 */
	Object context;

	/**
	 * The values of the {@link KTask#inputParameters}.
	 */
	Object[] inputValues;

	/**
	 * The values of the {@link KTask#outputParameters}.
	 */
	Object[] outputValues;

	Exception exception;

	private State state = State.CREATED;

	/**
	 * id of the procedure call
	 */
	private String id;

	protected KCall(KProcedure procedure, KProcess process) {
		this.procedure = procedure;
		this.process = process;
		this.inputValues = new Object[this.procedure.inputParameters.size()];
		this.outputValues = new Object[this.procedure.outputParameters.size()];
	}

	void enter() {
		this.state.enter(this);
	}

	public void start() {
		this.state.start(this);
	}

	public void complete() {
		this.state.complete(this);
	}

	public void cancel() {
		this.state.complete(this);
	}

	void fail() {
		this.state.fail(this);
	}
	
	void discard() {
		this.state.discard(this);
	}
	
	void succeed() {
		this.state.succeed(this);
	}
	
	

	public Runnable toRunnable() {
		return new Runnable() {
			@Override
			public void run() {
				KCall.this.start();
				KCall.this.complete();
			}
		};
	}

	protected void doStart() {
	}

	protected void doCancel() {
	}

	protected void doComplete() {
		this.procedure.complete(this);
	}

	protected void doEnter() {
		this.procedure.enter(this);
	}

	protected void doFail() {
		this.procedure.fail(this);
	}

	protected void doSucceed() {
		this.procedure.succeed(this);
	}

	protected void doDiscard() {
		// TODO Auto-generated method stub
		
	}

	void lock() {
		this.getProcess().lock();
	}

	void unlock() {
		this.getProcess().unlock();
	}

	protected void setState(State state) {
		this.state = state;
	}

	public Object getContext() {
		return context;
	}

	public void setContext(Object context) {
		this.context = context;
	}

	public KProcess getProcess() {
		return process;
	}

	public Object[] getInputValues() {
		return inputValues;
	}

	public Object[] getOutputValues() {
		return outputValues;
	}

	public Exception getException() {
		return exception;
	}

	protected void setException(Exception exception) {
		this.exception = exception;
	}

	State getState() {
		return state;
	}

	public String getId() {
		return id;
	}

	public KProcedure getProcedure() {
		return procedure;
	}

}