/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj.internal.kernel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.parallelj.internal.kernel.callback.Entry;
import org.parallelj.internal.kernel.callback.Exit;
import org.parallelj.internal.kernel.join.KAbstractJoin;

/**
 * TODO
 * 
 * <div class="note">
 * <p class="title">
 * Implementation note.
 * </p>
 * <p>
 * There is a strong link between {@link KProcedure} and {@link KProgram}.
 * Indeed, {@link KProcedure} could have been developed as an inner class of
 * {@link KProgram}.
 * </p>
 * But doing so, entry a {@link KProcedure} to a new kind of {@link KProgram},
 * which corresponds to a sub class of {@link KProcedure}, would imply a
 * subclass of {@link KProgram}. </p> </div>
 * 
 * @author Atos Worldline
 */
public class KProcedure extends KElement {

	class Marking {

		Set<KCall> calls = new HashSet<KCall>();
	}

	/**
	 * The join of this procedure.
	 */
	KJoin join;

	/**
	 * The split of this procedure.
	 */
	KSplit split;

	/**
	 * 
	 */
	Entry entry = new Entry() {

		@Override
		public void enter(KCall call) {
		}
	};

	/**
	 * 
	 */
	Exit exit = new Exit() {

		@Override
		public void exit(KCall call) {
		}
	};

	/**
	 * Liveness of this procedure: count the number of {@link KCall calls}
	 * currently running.
	 */
	KCondition liveness;

	KHandler handler;

	/**
	 * List of input parameters.
	 */
	List<KInputParameter> inputParameters = new ArrayList<KInputParameter>();

	/**
	 * List of output parameters.
	 */
	List<KOutputParameter> outputParameters = new ArrayList<KOutputParameter>();

	/**
	 * id of the procedure
	 */
	private String id;

	/**
	 * The procedure type.
	 */
	private String type = "<undef>";

	public KProcedure(KProgram program) {
		super(program);
		this.program.addProcedure(this);
		this.liveness = new KCondition(program, (short) 0, (short) 1);
	}

	@Override
	protected void init(KProcess process) {
		this.setMarking(process, this.newMarking(process));
	}

	protected Marking newMarking(KProcess process) {
		return new Marking();
	}

	@Override
	protected Marking getMarking(KProcess process) {
		return (Marking) super.getMarking(process);
	}

	/**
	 * Verify if this procedure is enabled.
	 * 
	 * The {@link #liveness} of this procedure and the
	 * {@link KProgram#verifyLiveness() liveness}of the enclosing program are
	 * verified as well as the {@link KAbstractJoin#isEnabled() enabling} of the
	 * join.
	 * 
	 * @return <code>true</code> if this procedure is enabled.
	 */
	public final boolean isEnabled(KProcess process) {
		return this.verifyLiveness(process) && this.join.isEnabled(process);
	}

	/**
	 * Verify if new {@link KCall} can be created according to {@link #liveness}
	 * capacity.
	 * 
	 * @param process
	 * @return
	 */
	boolean verifyLiveness(KProcess process) {
		return (this.liveness.size(process) < this.liveness.getCapacity());
	}

	void incrementLiveness(KProcess process) {
		this.program.liveness.produce(process);
		this.liveness.produce(process);
	}

	void decrementLiveness(KProcess process) {
		this.program.liveness.consume(process);
		this.liveness.consume(process);
	}

	final boolean tryFiring(KProcess process) {
		if (!this.isEnabled(process)) {
			return false;
		}
		KCall call = newCall(process);
		this.enter(call);
		return true;
	}

	final void enter(KCall call) {
		this.incrementLiveness(call.getProcess());
		this.join.join(call);
		this.getCalls(call.getProcess()).add(call);
		this.entry.enter(call);
	}

	void complete(KCall call) {
		if (!this.getCalls(call.getProcess()).remove(call)) {
			return ;//State.DISCARDED;
		}

		this.decrementLiveness(call.getProcess());
		call.getProcess().addCompleted(call);
	}

	void fail(KCall call) {

	}

	void succeed(KCall call) {
		this.exit.exit(call);
		if (this.split != null) {
			this.split.split(call);
		}
	}

	/**
	 * 
	 * @return
	 */
	protected KCall newCall(KProcess process) {
		return new KCall(this, process);
	}

	protected final void terminate(KProcess process) {
		Collection<KCall> collection = new ArrayList<KCall>(this
				.getCalls(process));
		for (KCall call : collection) {
			this.terminate(call);
		}
	}

	protected final void abort(KProcess process) {
		Collection<KCall> collection = new ArrayList<KCall>(this
				.getCalls(process));
		for (KCall call : collection) {
			this.abort(call);
		}

		// this.activity.liveness -= this.steps.size();
		// TODO program.getLiveness().decrement();
		// this.getInstances(program).clear();
	}

	protected void terminate(KCall call) {
	}

	protected void abort(KCall call) {
	}

	public boolean isBusy(KProcess process) {
		return this.liveness.contains(process);
	}

	/**
	 * @return the instances
	 */
	protected final Set<KCall> getCalls(KProcess process) {
		return this.getMarking(process).calls;
	}

	/**
	 * @return the join
	 */
	public KJoin getJoin() {
		return this.join;
	}

	/**
	 * @return the split
	 */
	public KSplit getSplit() {
		return this.split;
	}

	/**
	 * @param join
	 *            the join to set
	 */
	public void setJoin(KJoin join) {
		this.join = join;
	}

	/**
	 * @return
	 * @see KCondition#getCapacity()
	 */
	public short getCapacity() {
		return liveness.getCapacity();
	}

	/**
	 * @param capacity
	 * @see KCondition#setCapacity(short)
	 */
	public void setCapacity(short capacity) {
		liveness.setCapacity(capacity);
	}

	/**
	 * @param split
	 *            the split to set
	 */
	public void setSplit(KSplit split) {
		this.split = split;
	}

	public Entry getEntry() {
		return entry;
	}

	public void setEntry(Entry entry) {
		this.entry = entry;
	}

	public void addInputParameter(KInputParameter parameter) {
		this.inputParameters.add(parameter);
	}

	public void addOutputParameter(KOutputParameter parameter) {
		this.outputParameters.add(parameter);
	}

	public Exit getExit() {
		return exit;
	}

	public void setExit(Exit exit) {
		this.exit = exit;
	}

	public List<KOutputParameter> getOutputParameters() {
		return outputParameters;
	}

	public String getId() {
		return id;
	}

	public String getType() {
		return type;
	}

	public void setType(String type) {
		this.type = type;
	}

	public KHandler getHandler() {
		return handler;
	}

	public void setHandler(KHandler handler) {
		this.handler = handler;
	}

}