/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj.internal.kernel;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Represents an instance of a {@link KProgram}.
 * 
 * It cannot be instantiated directly; {@link KProgram#newProcess(Object)} must
 * be used instead.
 * 
 * @author Atos Worldline
 */
public class KProcess {

	/**
	 * Represents the state of a definition.
	 * 
	 * @author Atos Worldline
	 */
	public enum State {

		/**
		 * This is the initial state of a {@link KProcess}.
		 * 
		 * Transition to {@link #RUNNING} when {@link KProcess#start()()} is
		 * triggered.
		 * 
		 */
		PENDING {

			@Override
			void start(KProcess process) {
				process.doStart();
				process.setState(RUNNING);
			};

		},

		/**
		 * The definition is currently RUNNING meaning that some statements are
		 * currently RUNNING or new statements will be launched.
		 * 
		 * <p>
		 * Transition to {@link #ITERATING} when {@link KProcess#iterate()} is
		 * triggered.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #SUSPENDED} when {@link KProcess#suspend()} is
		 * triggered.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #COMPLETED} when all statements are complete.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #ABORTING} when {@link KProcess#abort()} is
		 * triggered.
		 * </p>
		 * 
		 * <p>
		 * Transition to {@link #TERMINATING} when
		 * {@link KDefinition#terminate()} is triggered.
		 * </p>
		 * 
		 */
		RUNNING {

			@Override
			Set<KCall> fire(KProcess process) {
				return process.doFire();
			};

			@Override
			void abort(KProcess process) {
				process.doAbort();
				process.setState(ABORTING);
			}

			@Override
			void terminate(KProcess process) {
				process.doTerminate();
				process.setState(TERMINATING);
			}

			@Override
			void complete(KProcess process) {
				process.doComplete();
				process.setState(State.COMPLETED);
			}

		},

		ABORTING {

			@Override
			void abort(KProcess process) {
			}

			@Override
			Set<KCall> fire(KProcess process) {
				return process.doFire();
			};

			@Override
			void complete(KProcess process) {
				process.doComplete();
				process.setState(State.ABORTED);
			}

		},

		TERMINATING {

			@Override
			void terminate(KProcess process) {
			}

			@Override
			Set<KCall> fire(KProcess process) {
				return process.doFire();
			};

			void complete(KProcess process) {
				process.doComplete();
				process.setState(State.TERMINATING);
			}

		},

		/**
		 * This is an exit condition of the process. No procedures are running
		 * and no new procedures can be launched.
		 */
		ABORTED,

		/**
		 * This is an exit condition of the process. No procedures are running
		 * and no new procedures can be launched.
		 */
		TERMINATED,

		/**
		 * This is the default exit condition of the process. The process has
		 * completed its work without interruption (cf. {@link KProcess#abort()}
		 * or {@link KProcess#terminate()}).
		 */
		COMPLETED;

		void start(KProcess process) {
			throw new IllegalStateException("TODO");
		}

		Set<KCall> fire(KProcess process) {
			throw new IllegalStateException("TODO");
		}

		void terminate(KProcess process) {
			throw new IllegalStateException("TODO");
		}

		void abort(KProcess process) {
			throw new IllegalStateException("TODO");
		}

		void complete(KProcess process) {
			throw new IllegalStateException("TODO");
		}
		
		/**
		 * Return <code>true</code> if this state is a final state.
		 * 
		 * @return
		 */
		public boolean isFinal() {
			return this == COMPLETED || this == ABORTED || this == TERMINATED;
		}

	}

	/**
	 * 
	 */
	private final KProgram program;

	/**
	 * KDefinition level context
	 */
	final Object context;

	/**
	 * Marking of all elements.
	 */
	Object[] markings;

	/**
	 * Lock used to protect methods such as {@link #iterate()}.
	 */
	private Lock lock = new ReentrantLock();

	/**
	 * The state of this process.
	 */
	private State state = State.PENDING;

	List<KCall> completed = new ArrayList<KCall>();

	/**
	 * id of the process
	 */
	private String id;

	/**
	 * The parent id.
	 */
	private String parentId = "0.0.0.0";

	protected KProcess(KProgram program, Object context) {
		this.program = program;
		this.context = context;
	}

	protected void setState(State state) {
		this.state = state;
	}

	void lock() {
		this.lock.lock();
	}

	void unlock() {
		this.lock.unlock();
	}

	boolean isEnabled() {
		return this.verify() && this.verifyLiveness();
	}

	boolean verifyLiveness() {
		return (this.program.liveness.size(this) < this.program.liveness
				.getCapacity());
	}

	void incrementLiveness() {
		this.program.liveness.produce(this);
	}

	void decrementLiveness() {
		this.program.liveness.consume(this);
	}

	boolean isBusy() {
		return this.program.liveness.size(this) != 0;
	}

	/**
	 * Verify a pre condition which is global to the definition. It will be
	 * called prior to try firing a execution (
	 * {@link KTask#tryFiring(KDefinition)}.
	 */
	public boolean verify() {
		return true;
	}

	/**
	 * Start this process.
	 */
	public final void start() {
		this.state.start(this);
	}

	/**
	 * Fire the enable procedures.
	 * 
	 * @return the procedure calls that are ready to be executed.
	 */
	public final Set<KCall> fire() {
		return this.state.fire(this);
	}

	public final void complete() {
		this.state.complete(this);
	}

	public final void terminate() {
		this.state.terminate(this);
	}

	public final void abort() {
		this.state.abort(this);
	}

	private void doStart() {
	}

	protected Set<KCall> doFire() {
		// handle completed calls prior starting new ones
		this.handleCompletedCalls();

		Set<KCall> set = new HashSet<KCall>();
		if (this.state == State.RUNNING) {
			// try to launch as much sub tasks as possible.
			for (KProcedure procedure : this.program.procedures) {
				// check pre condition
				if (!this.isEnabled()) {
					break;
				}
				while (this.isEnabled() && procedure.isEnabled(this)) {
					KCall call = procedure.newCall(this);
					call.enter();
					set.add(call);
				}
			}
		}

		// check completeness
		if (!this.isBusy()) {
			this.complete();
		}
		return set;

	}

	protected void doAbort() {
	}

	protected void doComplete() {
	}

	protected void doTerminate() {
	}

	void handleCompletedCalls() {
		switch (this.state) {
		case RUNNING:
			for (KCall call : this.completed) {
				if (call.getException() == null) {
					call.succeed();
				} else {
					if (call.getProcedure().getHandler() != null) {
						call.getProcedure().getHandler().addCallOnError(call);
						call.fail();
					} else {
						switch (this.program.policy) {
						case RESUME:
							call.succeed();
							break;
						case TERMINATE:
							call.discard();
							this.terminate();
							break;
						case ABORT:
							call.discard();
							this.abort();
							break;
						}
					}
				}
			}
			break;
		case TERMINATING:
			for (KCall call : this.completed) {
				if (call.getException() == null) {
					call.succeed();
				} else {
					call.discard();
				}
			}
			break;
		case ABORTING:
			for (KCall call : this.completed) {
				call.discard();
			}
		}
		this.completed.clear();

	}

	void addCompleted(KCall call) {
		this.completed.add(call);
	}

	public Object getContext() {
		return context;
	}

	public State getState() {
		return state;
	}

	public String getId() {
		return id;
	}

	public KProgram getProgram() {
		return program;
	}

	public String getParentId() {
		return parentId;
	}

	public void setParentId(String parentId) {
		this.parentId = parentId;
	}
}