/*
 *     ParallelJ, framework for parallel computing
 *
 *     Copyright (C) 2010 Atos Worldline or third-party contributors as
 *     indicated by the @author tags or express copyright attribution
 *     statements applied by the authors.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package org.parallelj.internal.kernel;

import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.parallelj.internal.kernel.KProcess.State;
import org.parallelj.internal.util.KCalls;
import org.parallelj.internal.util.StateEvent;
import org.parallelj.internal.util.StateListener;

/**
 * A processor is responsible to execute a {@link KProcess process} and all its
 * procedure {@link KCall calls}.
 * @author Atos Worldline
 * 
 */
public class KProcessor {

	/**
	 * stores the current processor; if any.
	 * 
	 */
	private static ThreadLocal<KProcessor> current = new ThreadLocal<KProcessor>();

	/**
	 * Sequence for processor ids.
	 */
	private static AtomicLong sequence = new AtomicLong(1);

	/**
	 * Identifier of this process.
	 */
	private long id = sequence.getAndIncrement();

	/**
	 * This class manage the execution of a process. It loops on procedure calls
	 * to execute and listen to their completion in order to execute the next
	 * procedure calls.
	 * 
	 */
	class Runner implements Runnable,
			StateListener<KCall, org.parallelj.internal.kernel.KCall.State> {

		/**
		 * Lock that will manage the concurrency access to this runner by the
		 * thread that loop on the process and the threads that inform the
		 * completion of calls.
		 */
		Lock lock = new ReentrantLock();

		/**
		 * The process to execute.
		 */
		KProcess process;

		/**
		 * Flag indicating that this runner is in the executor service.
		 */
		boolean active = false;

		Runner(KProcess process) {
			this.process = process;
		}

		@Override
		public void run() {
			if (this.execute() != 0) {
				this.activate();
			}
		}

		/**
		 * Execute the process: launch the enabled procedures.
		 * 
		 * @return the number of procedure calls submitted
		 */
		int execute() {
			this.lock.lock();

			// the number of calls submitted
			int count = 0;
			try {
				// start the process if not already started
				if (this.process.getState() == State.PENDING) {
					this.process.start();
				}

				// process is ended ?
				if (this.process.getState().isFinal()) {
					return 0;
				}
				// submit all procedure calls
				for (KCall call : this.process.fire()) {
					count++;
					KCalls.addStateListener(call, this);
					KProcessor.this.submit(call.toRunnable());
				}
			} finally {
				this.active = false;
				this.lock.unlock();
			}
			return count;
		}

		@Override
		public void stateChanged(
				StateEvent<KCall, org.parallelj.internal.kernel.KCall.State> event) {
			// a procedure call is completed, loop again.
			switch (event.getState()) {
			case COMPLETED:
			case FAILED:
			case DISCARDED:
			case SUCCEEDED:
				this.activate();
			}
		}

		void activate() {
			this.lock.lock();
			try {
				if (!this.active) {
					this.active = true;
					KProcessor.this.submit(this);
				}
			} finally {
				this.lock.unlock();
			}
		}
	}

	/**
	 * Executor that will execute the runnable
	 */
	private Executor executor;

	public KProcessor() {
		this(null);
	}

	public KProcessor(ExecutorService executor) {

		// if the executor is null, create a dummy one that just run the command
		this.executor = (executor != null) ? executor : new Executor() {

			@Override
			public void execute(Runnable command) {
				command.run();
			}
		};
	}

	/**
	 * 
	 * 
	 * @param process
	 */
	public void submit(KProcess process) {
		new Runner(process).activate();
	}

	/**
	 * Submit a runnable to be executed by the executor
	 * 
	 * @param runnable
	 */
	public void submit(final Runnable runnable) {
		this.executor.execute(new Runnable() {

			@Override
			public void run() {
				try {
					// install this processor in the thread local
					KProcessor.current.set(KProcessor.this);
					runnable.run();
				} finally {
					// uninstall this processor in the thread local
					KProcessor.current.set(null);
				}
			}
		});
	}

	public static KProcessor currentProcessor() {
		return current.get();
	}

	public long getId() {
		return id;
	}

}
